// kinem.h - Marcin Abram
// Juelich, Germany, 13.08.2009 r.
#include<cstdio>
#include<cstdlib>
#include<iostream>
#include<fstream>
#include<cmath>
#include<time.h>

///////////////////////////////////////////////////////
// PREPROCESOR ///////////////////////////////////////
/////////////////////////////////////////////////////
#define PI 3.1415926535
#define EPSILON 1e-12

///////////////////////////////////////////////////////
// DEKLARACJE KLAS ///////////////////////////////////
/////////////////////////////////////////////////////
class Ped;
class Pchniecie;

///////////////////////////////////////////////////////
// DEKLARACJE FUNKCJI ////////////////////////////////
/////////////////////////////////////////////////////
double Sqrt(double x);
double Max(double *m, unsigned N);
double Min(double *m, unsigned N);
double Sum(double *m, unsigned s, unsigned k);
double MaxKat(Ped pa, Ped pb, double *m, unsigned N);
double PrawKat(Ped pa, Ped pb, double* m, unsigned N, double alpha, unsigned K, unsigned histo=0);
double Kat(Ped p, Ped q);
void Sortuj(double *tab, unsigned N);
void Quicksort_base (double* tab, int left, int right);
double CalkaI(double b, double c, double x1, double x2);

///////////////////////////////////////////////////////
// DEKLARACJE ROZSZERZONE - PCHNIECIE ////////////////
/////////////////////////////////////////////////////
class Pchniecie
{
	private:
		double M_[4][4];

	public:
		Pchniecie(double gamma, double x_, double y_, double z_);
		Pchniecie(Ped pa, Ped pb);
		~Pchniecie(){}

		// wypisanie
		double gamma() { return M_[0][0]; }
		double beta() { return sqrt(1 - 1/(M_[0][0]*M_[0][0])); }
		void wypisz();

		// przeciazone operatory
		double& operator() (unsigned i, unsigned j);
		double operator() (unsigned i, unsigned j) const;
		Ped operator*(const Ped &p) const;
};

///////////////////////////////////////////////////////
// DEKLARACJE ROZSZERZONE - PED //////////////////////
/////////////////////////////////////////////////////
class Ped
{
	private:
		double p_[5];

	public:
		Ped();
		Ped(double E, double px, double py, double pz);
		Ped(double E, double m, double x, double y, double z);
		~Ped(){}

		// wypisz
		double E() const { return p_[0]; }
		double px() const { return p_[1]; }
		double py() const { return p_[2]; }
		double pz() const { return p_[3]; }
		double pp() const { return p_[4]; }
		double p() const { return Sqrt(p_[4]); }
		double mm() const { return (p_[0]*p_[0] - p_[4]); }
		double m() const { return Sqrt(p_[0]*p_[0] - p_[4]); }
		double ax() const { return p_[1]/Sqrt(p_[4]); }
		double ay() const { return p_[2]/Sqrt(p_[4]); }
		double az() const { return p_[3]/Sqrt(p_[4]); }
		
		void wypisz() { printf("(%8.3e, %8.3e, %8.3e, %8.3e)\n", p_[0], p_[1], p_[2], p_[3]); }
		void losuj(double E, double m);

		// przeciazanie operatorow
		int operator==(const Ped &p2) const;
		Ped operator+(const Ped &p2) const;
		Ped operator-(const Ped &p2) const;
		Ped operator*(const Pchniecie &M) const;
		double operator*(const Ped &p2) const;
		double operator() (unsigned i) const;
};

///////////////////////////////////////////////////////
// DEFINICJA - PCHNIECIA /////////////////////////////
/////////////////////////////////////////////////////
inline
Pchniecie::Pchniecie(double gamma, double x_, double y_, double z_)
{
	if(gamma < 1) { std::cerr << "Zla wartosc gammy!" << std::endl; } 
	// normalizuje wektor [x,y,z];
	double x[4] = {1, x_, y_, z_};
	double d = sqrt(x_*x_ + y_*y_ + z_*z_); for(int i=1; i<4; i++) { x[i] /= d; }
	// teraz mogo policzyc beta_calkowita, beta_x, beta_y, beta_z
	double beta[4]; for(int i=0; i<4; i++) { beta[i] = x[i]*sqrt(1 - 1/(gamma*gamma)); }
	// wypelnianie macierzy pchniecia
	for(int i=1; i<4; i++) for(int j=1; j<4; j++)
	{
		if(beta[0] != 0) M_[i][j]=(gamma-1)*(beta[i]/beta[0])*(beta[j]/beta[0]);
		else M_[i][j] = 0;
	}
	for(int i=1; i<4; i++) { M_[0][i]=M_[i][0]=beta[i]*gamma; M_[i][i]+=1; }
	M_[0][0] = gamma;
}

inline
Pchniecie::Pchniecie(Ped pa, Ped pb)
{
	Ped w = pa+pb; 
	// normalizuje wektor [x,y,z];
	double x[4] = {1, w.px(), w.py(), w.pz()};
	double d = sqrt(w.pp()); for(int i=1; i<4; i++) { x[i] /= d; }
	// teraz mogo policzyc beta_calkowita, beta_x, beta_y, beta_z oraz gamme
	double beta[4]; for(int i=0; i<4; i++) { beta[i] = x[i]*sqrt(w.pp())/w.E(); }
	double gamma = 1/sqrt(1-beta[0]*beta[0]);
	// wypelnianie macierzy pchniecia
	for(int i=1; i<4; i++) for(int j=1; j<4; j++)
	{
		if(beta[0] != 0) M_[i][j]=(gamma-1)*(beta[i]/beta[0])*(beta[j]/beta[0]);
		else M_[i][j] = 0;
	}
	for(int i=1; i<4; i++) { M_[0][i]=M_[i][0]=-beta[i]*gamma; M_[i][i]+=1; }
	M_[0][0] = gamma;
}

inline
void Pchniecie::wypisz()
{
	for(int i=0; i<4; i++)
	{
		for(int j=0; j<4; j++) { printf("%9.3f ", M_[i][j]); }
		printf("\n");
	}
}

inline
double& Pchniecie::operator() (unsigned i, unsigned j)
{
	if (i >= 4 || i >= 4)
	{ std::cerr << "Wykroczenie poza zakrez Pchniecia" << std::endl;
	  exit(EXIT_FAILURE); }
	return M_[i][j];
}

inline
double Pchniecie::operator() (unsigned i, unsigned j) const
{
	if (i >= 4 || i >= 4)
	{ std::cerr << "Wykroczenie poza zakrez Pchniecia" << std::endl;
	  exit(EXIT_FAILURE); }
	return M_[i][j];
}

inline
Ped Pchniecie::operator*(const Ped &p) const
{
	double t[4] = {0, 0, 0, 0};
	for(int i=0; i<4; i++) for(int j=0; j<4; j++) { t[i] += M_[i][j]*p(j); }
	return Ped(t[0], t[1], t[2], t[3]);
}

///////////////////////////////////////////////////////
// DEFINICJA - PEDU //////////////////////////////////
/////////////////////////////////////////////////////
inline
Ped::Ped()
{
	for(int i=0; i<5; i++) { p_[i] = 0;  }
}

inline
Ped::Ped(double E, double px, double py, double pz)
{
	p_[0] = E; p_[1] = px; p_[2] = py; p_[3] = pz;
	p_[4] = px*px + py*py + pz*pz;
	if(p_[0]*p_[0] < p_[4]*(1-EPSILON))
	{ std::cerr << "Pojawil sie czteroped czastki o ujemnej masie! (1)" << std::endl; }
}

Ped::Ped(double E, double m, double x, double y, double z)
{
	// normalizacja x, y, z
	double d=x*x+y*y+z*z;
	if(d==0 && E==m) { p_[0]=E, p_[1]=p_[2]=p_[3]=p_[4]=0; }
	else if(d==0 && E!=m)
	{
		std::cerr << "Dla czastki spoczywajacej powinno byc E=m" <<std::endl;
		exit(EXIT_FAILURE);
	}
	else
	{
		double ax=x/d, ay=y/d, az=z/d;
		p_[0] = E; p_[4] = E*E - m*m;
		if(p_[0]*p_[0] < p_[4]*(1-EPSILON))
	{
		std::cerr << "Pojawil sie czteroped czastki o ujemnej masie! (2)" << std::endl;
		exit(EXIT_FAILURE);
	}
		double p=sqrt(p_[4]); p_[1] = p*ax; p_[2] = p*ay; p_[3] = p*az;
	}
}

void Ped::losuj(double E, double m)
{
	double phi = static_cast<double>(rand()%(1000000))/1000000*2*PI;
	double theta = static_cast<double>(rand()%(1000000))/1000000*PI;
	double p = sqrt(E*E - m*m);
	p_[0]=E;
	p_[1]=p*sin(theta)*cos(phi);
	p_[2]=p*sin(theta)*sin(phi);
	p_[3]=p*cos(theta);
	p_[4]=E*E-m*m;
}

inline
int Ped::operator==(const Ped &p2) const
{
	return (this->mm() == p2.mm()*(1-EPSILON)) && \
	(this->mm() <= p2.mm() + EPSILON);
};

inline
Ped Ped::operator+(const Ped &p2) const
{ return Ped(p_[0]+p2.p_[0], p_[1]+p2.p_[1], p_[2]+p2.p_[2], p_[3]+p2.p_[3]); }

inline
Ped Ped::operator-(const Ped &p2) const
{ return Ped(p_[0]-p2.p_[0], p_[1]-p2.p_[1], p_[2]-p2.p_[2], p_[3]-p2.p_[3]); }

inline
Ped Ped::operator*(const Pchniecie &M) const
{
	double t[4] = {0, 0, 0, 0}, n[4] = {-1, 1, 1, 1};
	for(int i=0; i<4; i++) for(int j=0; j<4; j++) { t[i] += M(i,j)*p_[j]*n[i]*n[j]; }
	return Ped(t[0], t[1], t[2], t[3]);
}

inline
double Ped::operator*(const Ped &p2) const
{
	double iloczyn=0;
	for(unsigned i=1; i<4; i++) { iloczyn += p_[i]*p2.p_[i]; }
	return iloczyn;
}


inline
double Ped::operator() (unsigned i) const
{
	if (i >= 4)
	{ std::cerr << "Wykroczenie poza zakrez Pchniecia" << std::endl; exit(EXIT_FAILURE); }
	return p_[i];
}

///////////////////////////////////////////////////////
// DEFINICJA - INNE FUNKCJE //////////////////////////
/////////////////////////////////////////////////////
inline
double Sqrt(double x)
{
	if(x<0) x=-x;
	return sqrt(x);
}
// koniec Sqrt

double Max(double *m, unsigned N)
{
	double t=m[0];
	for(unsigned i=1; i<N; i++) { if(t<m[i]) t=m[i]; }
	return t;
}
// koniec Max

double Min(double *m, unsigned N)
{
	double t=m[0];
	for(unsigned i=1; i<N; i++) { if(t>m[i]) t=m[i]; }
	return t;
}
// koniec Max

double Sum(double *m, unsigned s, unsigned k)
{
	double t=0;
	for(unsigned i=s; i<k; i++) { t += m[i]; }
	return t;
}
// koenic Sum

double MaxKat(Ped pa, Ped pb, double *m, unsigned N)
{
	// badam przypadek graniczny, kiedy to najlezejsza i najciezsza czastka z grupy
	// bedzie sie poruszac, a reszta bedzie stac w miejscu (w CM).
	// wszystko z min odnosi się do najlzejszej czastki, z max do najciezszej
	// NIE OZNACZA TO, ze mamy MINIMUM ENERGI lub PEDU
	
	// LE to energia do dyspozycji dla czastek maxm i minx
	Pchniecie L(pa, pb);
	Ped Lpa=L*pa, Lpb=L*pb;
	double mmax	= Max(m, N), mmin = Min(m, N);
	double LE = Lpa.E() + Lpb.E() - Sum(m, 0, N) + mmax + mmin;
	if(LE < mmax + mmin)
	{
		printf("Czastki A i B maja za mala energie, zeby wytworzyc rzadane czastki\n");
		exit(EXIT_SUCCESS);
	}
	double LEmin = (mmin*mmin - mmax*mmax + LE*LE)/(2*LE);
	double Lpmin = sqrt(LEmin*LEmin - mmin*mmin);
	double v = Lpmin/LEmin;

	// patrz ksiazka: "Relativistic Kinematics", Hagedorn, Benjamin (1964) srt. ~50
	if(v > L.beta()) { return PI; }
	//else if( v <= L.beta()*(1+EPSILON) && v >= L.beta()*(1-EPSILON)) { return PI/2; }
	else if( v == L.beta()) { return PI/2; }
	else
	{
		double tgTheta2 = ((v*v)/((L.beta() - v)*(L.beta() + v)))/(L.gamma()*L.gamma());
		return atan(sqrt(tgTheta2));
	}
}
// koniec MaxKat

double PrawKat(Ped pa, Ped pb, double* m, unsigned N, double alpha, unsigned K, unsigned histo)
{
	Pchniecie L(pa, pb);
	Ped Lpa=L*pa, Lpb=L*pb;
	double mmax	= Max(m, N), mmin = Min(m, N);
	double LE = Lpa.E() + Lpb.E() - Sum(m, 0, N) + mmax + mmin;
	if(LE < mmax + mmin)
	{
		printf("Czastki A i B maja za mala energie, zeby wytworzyc rzadane czastki\n");
		exit(EXIT_SUCCESS);
	}
	double LEmin = (mmin*mmin - mmax*mmax + LE*LE)/(2*LE);

	Ped Lq;
	double *katy = new double[K];
	srand(time(0));
	for(unsigned i=0; i<K; i++)
	{
		Lq.losuj(LEmin, mmin);
		Ped q = Lq*L;
		katy[i] = Kat(pa+pb, q);
	}
	Sortuj(katy, K);
	
	double wynik;
	if(alpha > 1 || alpha < 0)
	{ std::cerr << "Niedozwolony zakres aplha" << std::endl; }
	else if(alpha == 1)
	{ wynik = katy[K-1]; }
	else
	{ wynik = katy[static_cast<unsigned>(K*alpha)]; }
	
	if(histo)
	{
		std::ofstream os;
		os.open("hist2");
		for(double  d=0; d<1; d += 0.001)
		{ os << d << " " << katy[static_cast<unsigned>(K*d)] << std::endl; }
		os.close(); 											 
	}

	delete[] katy;
	return wynik;
}
// koniec PrawKat

inline
double Kat(Ped p, Ped q) { return 180*acos((p*q)/(p.p()*q.p()))/PI; }
// koniec Kat

void Sortuj(double *tab, unsigned N)
{
	// Uzywam Quicksort
	Quicksort_base(tab, 0, static_cast<int>(N));
}
// koniec Sortuj

void Quicksort_base (double* tab, int left, int right)
{
	if (left < right)
	{
		int medium=left;
		for(int i=left+1; i <= right; i++)
		{
			if (tab[i] < tab[left])
			{
				medium++;
				double temp=tab[medium]; tab[medium]=tab[i]; tab[i]=temp;
			}
		}
		double temp=tab[left]; tab[left]=tab[medium]; tab[medium]=temp;
		Quicksort_base(tab, left, medium-1);
		Quicksort_base(tab, medium+1, right);
	}
}
// koniec Quicksort_base

double CalkaI(double b, double c, double x1, double x2)
{
	double I1 = (-b+x1)*(b*b+c-2*b*x1+x1*x1)/(2*sqrt(b*b+c-2*b*x1+x1*x1)) \
		- c/2 * log(2*(b + sqrt(c+(b-x1)*(b-x1))) - x1);
	double I2 = (-b+x2)*(b*b+c-2*b*x2+x2*x2)/(2*sqrt(b*b+c-2*b*x2+x2*x2)) \
		- c/2 * log(2*(b + sqrt(c+(b-x2)*(b-x2))) - x2);
	return I2-I1;
}
// koniec CalkaI
